#!/usr/bin/python
# -*- coding: utf-8 -*-

# SàT: a XMPP client
# Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014 Jérôme Poisson (goffi@goffi.org)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""High level logging functions"""
# XXX: this module use standard logging module when possible, but as SàT can work in different cases where logging is not the best choice (twisted, pyjamas, etc), it is necessary to have a dedicated module. Additional feature like environment variables and colors are also managed.

from sat.core.constants import Const as C
from sat.core import exceptions

_backend = None
_loggers = {}
_handlers = {}


class Logger(object):
    """ High level logging class """

    def __init__(self, name):
        self._name = name

    def debug(self, msg):
        print msg

    def info(self, msg):
        print msg

    def warning(self, msg):
        print msg

    def error(self, msg):
        print msg

    def critical(self, msg):
        print msg


class FilterName(object):

    def __init__(self, name_re):
        """Initialise name filter

        @param name_re: regular expression used to filter names (using search and not match)
        """
        assert name_re
        import re
        self.name_re = re.compile(name_re)

    def filter(self, record):
        if self.name_re.search(record.name) is not None:
            return 1
        return 0

def _manageOutputs(outputs_raw):
    """ Parse output option in a backend agnostic way, and fill _backend consequently

    @param outputs_raw: output option as enterred in environment variable or in configuration
    """
    if not outputs_raw:
        return
    outputs = outputs_raw.split(C.LOG_OPT_OUTPUT_SEP)
    global _handlers
    if len(outputs) == 1:
        _handlers[C.LOG_OPT_OUTPUT_FILE] = outputs.pop()

    for output in outputs:
        if not output:
            continue
        if output[-1] == ')':
            # we have options
            opt_begin = output.rfind('(')
            options = output[opt_begin+1:-1]
            output = output[:opt_begin]
        else:
            options = None

        if output not in (C.LOG_OPT_OUTPUT_DEFAULT, C.LOG_OPT_OUTPUT_FILE, C.LOG_OPT_OUTPUT_MEMORY):
            raise ValueError(u"Invalid output [%s]" % output)

        if output == C.LOG_OPT_OUTPUT_DEFAULT:
            # no option for defaut handler
            _handlers[output] = None
        elif output == C.LOG_OPT_OUTPUT_FILE:
            if not options:
                ValueError("%(handler)s output need a path as option" % {'handler': output})
            _handlers[output] = options
            options = None # option are parsed, we can empty them
        elif output == C.LOG_OPT_OUTPUT_MEMORY:
            # we have memory handler, option can be the len limit or None
            try:
                limit = int(options)
                options = None # option are parsed, we can empty them
            except (TypeError, ValueError):
                limit = C.LOG_OPT_OUTPUT_MEMORY_LIMIT
            _handlers[output] = limit

        if options: # we should not have unparsed options
            raise ValueError(u"options [%(options)s] are not supported for %(handler)s output" % {'options': options, 'handler': output})

def _configureStdLogging(logging, level=None, fmt=C.LOG_OPT_FORMAT[1], output=C.LOG_OPT_OUTPUT[1], logger=None, colors=False, force_colors=False):
    """Configure standard logging module

    @param logging: standard logging module
    @param level: one of C.LOG_LEVELS
    @param fmt: format string, pretty much as in std logging. Accept the following keywords (maybe more depending on backend):
        - "message"
        - "levelname"
        - "name" (logger name)
    @param logger: if set, use it as a regular expression to filter on logger name.
        Use search to match expression, so ^ or $ can be necessary.
    @param colors: if True use ANSI colors to show log levels
    @param force_colors: if True ANSI colors are used even if stdout is not a tty
    """
    format_ = fmt
    if level is None:
        level = C.LOG_LVL_DEBUG
    import sys
    with_color = colors & (sys.stdout.isatty() or force_colors)
    if not colors and force_colors:
        raise ValueError("force_colors can't be used if colors is False")

    class SatFormatter(logging.Formatter):
        u"""Formatter which manage SàT specificities"""

        def __init__(self, fmt=None, datefmt=None):
            super(SatFormatter, self).__init__(fmt, datefmt)

        def format(self, record):
            s = super(SatFormatter, self).format(record)
            if with_color:
                if record.levelno == logging.DEBUG:
                    fmt = (C.ANSI_FG_CYAN, s, C.ANSI_RESET)
                elif record.levelno == logging.WARNING:
                    fmt = (C.ANSI_FG_YELLOW, s, C.ANSI_RESET)
                elif record.levelno == logging.ERROR:
                    fmt = (C.ANSI_FG_RED,
                           C.ANSI_BLINK,
                           r'/!\ ',
                           C.ANSI_BLINK_OFF,
                           s,
                           C.ANSI_RESET)
                elif record.levelno == logging.CRITICAL:
                    fmt = (C.ANSI_BOLD,
                           C.ANSI_FG_RED,
                           'Guru Meditation ',
                           C.ANSI_NORMAL_WEIGHT,
                           s,
                           C.ANSI_RESET)
                else:
                    fmt = s
                s = ''.join(fmt)

            return s

    root_logger = logging.getLogger()
    if len(root_logger.handlers) == 0:
        _manageOutputs(output)
        formatter = SatFormatter(format_)
        name_filter = FilterName(logger) if logger else None
        for handler, options in _handlers.items():
            if handler == C.LOG_OPT_OUTPUT_DEFAULT:
                hdlr = logging.StreamHandler()
            elif handler == C.LOG_OPT_OUTPUT_MEMORY:
                import logging.handlers
                hdlr = logging.handlers.BufferingHandler(options)
            elif handler == C.LOG_OPT_OUTPUT_FILE:
                import os.path
                hdlr = logging.FileHandler(os.path.expanduser(options))
            else:
                raise ValueError("Unknown handler type")
            hdlr.setFormatter(formatter)
            root_logger.addHandler(hdlr)
            root_logger.setLevel(level)
            if name_filter is not None:
                hdlr.addFilter(name_filter)
    else:
        root_logger.warning(u"Handlers already set on root logger")

def configure(backend=C.LOG_BACKEND_STANDARD, **options):
    """Configure logging bejaviour
    @param backend: can be:
        C.LOG_BACKEND_STANDARD: use standard logging module
        C.LOG_BACKEND_TWISTED: use twisted logging module (with standard logging observer)
        C.LOG_BACKEND_BASIC: use a basic print based logging
    """
    global _backend
    if _backend is not None:
        raise exceptions.InternalError("Logging can only be configured once")
    _backend = backend

    if backend in (C.LOG_BACKEND_TWISTED, C.LOG_BACKEND_STANDARD):
        if backend ==  C.LOG_BACKEND_TWISTED:
            from twisted.python import log
            observer = log.PythonLoggingObserver()
            observer.start()
        global getLogger
        global debug
        global info
        global warning
        global error
        global critical
        import logging
        _configureStdLogging(logging, **options)
        getLogger = logging.getLogger
        debug = logging.debug
        info = logging.info
        warning = logging.warning
        error = logging.error
        critical = logging.critical

    elif backend == C.LOG_BACKEND_BASIC:
        pass

    else:
        raise ValueError("unknown backend")

def _parseOptions(options):
    """Parse string options as given in conf or environment variable, and return expected python value

    @param options (dict): options with (key: name, value: string value)
    """
    COLORS = C.LOG_OPT_COLORS[0]
    LEVEL = C.LOG_OPT_LEVEL[0]

    if COLORS in options:
        if options[COLORS].lower() in ('1', 'true'):
            options[COLORS] = True
        elif options[COLORS] == 'force':
            options[COLORS] = True
            options['force_colors'] = True
        else:
            options[COLORS] = False
    if LEVEL in options:
        level = options[LEVEL].upper()
        if level not in C.LOG_LEVELS:
            level = C.LOG_LVL_INFO
        options[LEVEL] = level

def satConfigure(backend=C.LOG_BACKEND_TWISTED):
    """Configure logging system for SàT, can be used by frontends

    logs conf is read in SàT conf, then in environment variables. It must be done before Memory init
    """
    import ConfigParser
    import os
    log_conf = {}
    config = ConfigParser.SafeConfigParser()
    config.read(C.CONFIG_FILES)
    for opt_name, opt_default in C.LOG_OPTIONS:
        try:
            log_conf[opt_name] = os.environ[''.join((C.ENV_PREFIX, C.LOG_OPT_PREFIX.upper(), opt_name.upper()))]
        except KeyError:
            try:
                log_conf[opt_name] = config.get('DEFAULT', C.LOG_OPT_PREFIX + opt_name)
            except ConfigParser.NoOptionError:
                log_conf[opt_name] = opt_default

    _parseOptions(log_conf)
    configure(backend, **log_conf)

def getLogger(name=C.LOG_BASE_LOGGER):
    return _loggers.setdefault(name, Logger(name))

_root_logger = getLogger()

def debug(msg):
    _root_logger.debug(msg)

def info(msg):
    _root_logger.info(msg)

def warning(msg):
    _root_logger.warning(msg)

def error(msg):
    _root_logger.error(msg)

def critical(msg):
    _root_logger.critical(msg)
